import tempfile

from pydal import DAL
import uuid
import pytest
import dotmap

from src.edwh_auth_rbac.model import define_auth_rbac_model
from src.edwh_auth_rbac.rbac import AuthRbac
from src.edwh_auth_rbac.migrations import rbac_migrations

namespace = uuid.UUID('84f5c757-4be0-49c8-a3ba-4f4d79167839')


@pytest.fixture(scope="module")
def tmpdir():
    with tempfile.TemporaryDirectory() as tmpdirname:
        print('new tempdir')
        yield tmpdirname


@pytest.fixture(scope="module")
def database(tmpdir):
    class Database:
        def __enter__(self):
            self.db = DAL('sqlite://auth_rbac.sqlite', folder=tmpdir)

            define_auth_rbac_model(self.db, dict(allowed_types=['user', 'group']))
            rbac_migrations(self.db)
            return self.db

        def __exit__(self, exc_type, exc_value, traceback):
            self.db.close()

    return Database()


@pytest.fixture(scope="module")
def rbac(database):
    with database as db:
        yield AuthRbac(db)


@pytest.fixture(scope="module")
def store(_=dotmap.DotMap()):
    print('store', _)
    return _


@pytest.mark.incremental
class TestSequentially:
    def test_drop_all_test_users(self, database):
        with database as db:
            users = db(db.identity.email.contains('@test.nl')).select()
            db(db.identity.email.contains('@test.nl')).delete()
            for user in users:
                db((db.membership.member_of == user.object_id) | (db.membership.subject == user.object_id)).delete()
                db((db.permission.identity_object_id == user.object_id) | (
                        db.permission.target_object_id == user.object_id)).delete()
            db.commit()
            assert db(db.identity.email.contains('@test.nl')).count() == 0, 'Howcome @test.nl still exist?'

    def test_user_creation(self, rbac, store):
        store.remco = rbac.add_user('remco@test.nl', 'remco', 'remco test', 'secret', [])['object_id']
        store.pietje = rbac.add_user('pietje@test.nl', 'pietje', 'pietje test', 'secret', [])['object_id']
        store.truus = rbac.add_user('truus@test.nl', 'truus', 'truus test', 'secret', [])['object_id']

    def test_group_creation(self, rbac, store):
        store.articles = rbac.add_group('articles@test.nl', 'articles', [])['object_id']
        store.all = rbac.add_group('all@test.nl', 'all', [])['object_id']
        store.users = rbac.add_group('users@test.nl', 'users', [])['object_id']
        store.admins = rbac.add_group('admins@test.nl', 'admins', [])['object_id']

    def test_item_creation(self, rbac, store):
        for name in 'abcde':
            store[name] = rbac.add_user('article_' + name + '@test.nl', name, 'article', '', [])[
                'object_id']

    def test_stash_users_in_groups(self, rbac, store):
        rbac.add_membership(store.remco, store.admins)
        rbac.add_membership(store.pietje, store.users)
        rbac.add_membership(store.truus, store.users)
        rbac.add_membership(store.users, store.all)
        rbac.add_membership(store.admins, store.all)

    def test_stash_items_in_groups(self, rbac, store):
        for name in 'abcde':
            rbac.add_membership(store[name], store.articles)

    def test_add_some_permissions(self, rbac, store):
        rbac.add_permission(store.admins, store.articles, 'read')
        rbac.add_permission(store.admins, store.articles, 'write')
        rbac.add_permission(store.users, store.articles, 'read')

    def test_first_level_memberships(self, rbac, store):
        assert rbac.has_membership(store.remco, store.admins) is True
        assert rbac.has_membership(store.pietje, store.users) is True
        assert rbac.has_membership(store.remco, store.users) is False
        assert rbac.has_membership(store.pietje, store.admins) is False

    def test_second_level_memberships(self, rbac, store):
        assert rbac.has_membership(store.remco, store.all) is True
        assert rbac.has_membership(store.pietje, store.all) is True

    def test_first_level_permissions(self, rbac, store):
        assert rbac.has_permission(store.admins, store.articles, 'read') is True
        assert rbac.has_permission(store.admins, store.articles, 'write') is True
        assert rbac.has_permission(store.users, store.articles, 'read') is True
        assert rbac.has_permission(store.users, store.articles, 'write') is False

    def test_second_to_first_level_permissions(self, rbac, store):
        assert rbac.has_permission(store.remco, store.articles, 'read') is True
        assert rbac.has_permission(store.remco, store.articles, 'write') is True
        assert rbac.has_permission(store.pietje, store.articles, 'read') is True
        assert rbac.has_permission(store.pietje, store.articles, 'write') is False

    def test_second_to_second_level_permissions(self, rbac, store):
        assert rbac.has_permission(store.remco, store.a, 'read') is True
        assert rbac.has_permission(store.remco, store.a, 'write') is True
        assert rbac.has_permission(store.pietje, store.a, 'read') is True
        assert rbac.has_permission(store.pietje, store.a, 'write') is False

    def test_deeper_group_nesting(self, rbac, store):
        store.subadmins = rbac.add_group('sub_admins@test.nl', 'subadmins', [])['object_id']
        store.subarticles = rbac.add_group('sub_articles@test.nl', 'subarticles', [])['object_id']
        rbac.add_membership(store.subarticles, store.articles)
        rbac.add_membership(store.subadmins, store.admins)
        store.nested_admin = rbac.add_user('nested_admin@test.nl', 'nested_admin', 'nested_admin test', 'secret', [])[
            'object_id']
        rbac.add_membership(store.nested_admin, store.subadmins)
        for name in 'stuvw':
            store[name] = rbac.add_user('article_' + name + '@test.nl', name, 'subarticle', '', [])['object_id']
            rbac.add_membership(store[name], store.subarticles)
        assert rbac.has_permission(store.nested_admin, store.s, 'read') is True

    def test_removing_a_nested_group(self, rbac, store):
        rbac.remove_membership(store.nested_admin, store.subadmins)
        assert rbac.has_permission(store.nested_admin, store.s, 'read') is False



import copy
import datetime
import hashlib
import hmac
import uuid

import dateutil.parser
from pydal import Field
from pydal import SQLCustomType
from pydal.objects import SQLALL, Query

from .helpers import IS_IN_LIST


class DEFAULT:
    pass


DEFAULT_STARTS = datetime.datetime(2000, 1, 1)
DEFAULT_ENDS = datetime.datetime(3000, 1, 1)


def unstr_datetime(s):
    """json helper... might values arrive as str """
    return dateutil.parser.parse(s) if isinstance(s, str) else s


class Password:
    """
    Encode a password using: Password.encode('secret')
    """

    @classmethod
    def hmac_hash(cls, value, key, salt=None):
        digest_alg = hashlib.sha512
        d = hmac.new(str(key).encode(), str(value).encode(), digest_alg)
        if salt:
            d.update(str(salt).encode())
        return d.hexdigest()

    @classmethod
    def validate(cls, password, candidate):
        salt, hashed = candidate.split(':', 1)
        return cls.hmac_hash(value=password, key='secret_start', salt=salt) == hashed

    @classmethod
    def encode(cls, password):
        salt = uuid.uuid4().hex
        return salt + ':' + cls.hmac_hash(value=password, key='secret_start', salt=salt)


def is_uuid(s):
    try:
        uuid.UUID(s)
        return True
    except Exception:
        return False


def key_lookup_query(db, identity_key: str | int | uuid.UUID, object_type=None) -> Query:
    if '@' in str(identity_key):
        query = db.identity.email == identity_key.lower()
    elif isinstance(identity_key, int):
        query = db.identity.id == identity_key
    elif is_uuid(identity_key):
        query = db.identity.object_id == identity_key.lower()
    else:
        query = db.identity.firstname == identity_key

    if object_type:
        query &= db.identity.object_type == object_type

    return query


def key_lookup(db, identity_key, object_type=None) -> str | None:
    query = key_lookup_query(db, identity_key, object_type)

    rowset = db(query).select(db.identity.object_id)

    if not rowset:
        return None
    elif len(rowset) > 1:
        raise ValueError('Keep lookup for {} returned {} results.'.format(identity_key, len(rowset)))

    return rowset.first().object_id


my_datetime = SQLCustomType(type='string',
                            native='char(35)',
                            encoder=(lambda x: x.isoformat(' ')),
                            decoder=(lambda x: dateutil.parser.parse(x)))


def define_auth_rbac_model(db, other_args):
    db.define_table('identity',
                    # std uuid from uuid libs are 36 chars long
                    Field('object_id', 'string', length=36, unique=True, notnull=True, default=str(uuid.uuid4())),
                    Field('object_type', 'string', requires=(IS_IN_LIST(other_args['allowed_types']))),
                    Field('created', 'datetime', default=datetime.datetime.now),
                    # email needn't be unique, groups can share email addresses, and with people too
                    Field('email', 'string'),
                    Field('firstname', 'string', comment='also used as short code for groups'),
                    Field('fullname', 'string'),
                    Field('encoded_password', 'string'),
                    )

    db.define_table('membership',
                    # beide zijn eigenlijk: reference:identity.object_id
                    Field('subject', 'string', length=36, notnull=True),
                    Field('member_of', 'string', length=36, notnull=True),
                    # Field('starts','datetime', default=DEFAULT_STARTS),
                    # Field('ends','datetime', default=DEFAULT_ENDS),
                    )

    db.define_table('permission',
                    Field('privilege', 'string', length=20),
                    # reference:identity.object_id
                    Field('identity_object_id', 'string', length=36),
                    Field('target_object_id', 'string', length=36),
                    # Field('scope'), lets bail scope for now. every one needs a rule for everything
                    # just to make sure, no 'wildcards' and 'every dossier for org x' etc ...
                    Field('starts', type=my_datetime, default=DEFAULT_STARTS),
                    Field('ends', type=my_datetime, default=DEFAULT_ENDS),
                    )

    db.define_table('recursive_memberships',
                    Field('root'),
                    Field('object_id'),
                    Field('object_type'),
                    Field('level', 'integer'),
                    Field('email'),
                    Field('firstname'),
                    Field('fullname'),
                    migrate=False,
                    primarykey=['root', 'object_id']  # composed, no primary key
                    )
    db.define_table('recursive_members',
                    Field('root'),
                    Field('object_id'),
                    Field('object_type'),
                    Field('level', 'integer'),
                    Field('email'),
                    Field('firstname'),
                    Field('fullname'),
                    migrate=False,
                    primarykey=['root', 'object_id']  # composed, no primary key
                    )


def add_identity(db, email, password, member_of, name=None, firstname=None, fullname=None, gid=None, object_type=None):
    """paramaters name and firstname are equal. """
    email = email.lower().strip()
    if object_type is None:
        raise ValueError('object_type parameter expected')
    object_id = gid if gid else str(uuid.uuid4())
    db.identity.validate_and_insert(
        object_id=object_id,
        object_type=object_type,
        email=email,
        firstname=name or firstname or None,
        fullname=fullname,
        encoded_password=Password.encode(password)
    )
    db.commit()
    for key in member_of:
        group_id = key_lookup(db, key, 'group')
        if get_group(db, group_id):
            # check each group if it exists.
            add_membership(db, identity_key=object_id, group_key=group_id)
    db.commit()
    return object_id


def add_group(db, email, name, member_of):
    return add_identity(db, email, None, member_of, name=name, object_type='group')


def remove_identity(db, object_id):
    removed = db(db.identity.object_id == object_id).delete()
    # todo: remove permissions and group memberships
    db.commit()
    return removed > 0


def get_identity(db, key, object_type=None):
    """
    :param db: dal db connection
    :param key: can be the email, id, or object_id
    :return: user record or None when not found
    """
    query = key_lookup_query(db, key, object_type)
    rows = db(query).select()
    return rows.first()


def get_user(db, key):
    """
    :param db: dal db connection
    :param key: can be the email, id, or object_id
    :return: user record or None when not found
    """
    return get_identity(db, key, object_type='user')


def get_group(db, key):
    """

    :param db: dal db connection
    :param key: can be the name of the group, the id, object_id or email_address
    :return: user record or None when not found
    """
    return get_identity(db, key, object_type='group')


def authenticate_user(db, password=None, user=None, key=None):
    if not password:
        return False
    if not user:
        user = get_user(db, key)
    return Password.validate(password, user.encoded_password)


def add_membership(db, identity_key, group_key):
    identity_oid = key_lookup(db, identity_key)
    if identity_oid is None:
        raise ValueError('invalid identity_oid key: ' + identity_key)
    group = get_group(db, group_key)
    if not group:
        raise ValueError('invalid group key: ' + group_key)
    query = db.membership.subject == identity_oid
    query &= db.membership.member_of == group.object_id
    if db(query).count() == 0:
        db.membership.validate_and_insert(
            subject=identity_oid,
            member_of=group.object_id,
        )
    db.commit()


def remove_membership(db, identity_key, group_key):
    identity = get_identity(db, identity_key)
    group = get_group(db, group_key)
    query = db.membership.subject == identity.object_id
    query &= db.membership.member_of == group.object_id
    deleted = db(query).delete()
    db.commit()
    return deleted


def get_memberships(db, object_id, bare=True):
    query = db.recursive_memberships.root == object_id
    fields = [db.recursive_memberships.object_id, db.recursive_memberships.object_type]
    if not bare:
        fields = []
    return db(query).select(*fields)


def get_members(db, object_id, bare=True):
    query = db.recursive_members.root == object_id
    fields = [db.recursive_members.object_id, db.recursive_members.object_type]
    if not bare:
        fields = []
    return db(query).select(*fields)


def add_permission(db, identity_key, target_oid, privilege, starts=DEFAULT_STARTS, ends=DEFAULT_ENDS):
    identity_oid = key_lookup(db, identity_key)
    starts = unstr_datetime(starts)
    ends = unstr_datetime(ends)
    if has_permission(db, identity_oid, target_oid, privilege, when=starts):
        # permission already granted. just skip it
        print(
            '{privilege} permission already granted to {user_or_group_key} on {target_oid} @ {starts} '.format(
                **locals()))
        # print(db._lastsql)
        return
    db.permission.validate_and_insert(
        privilege=privilege,
        identity_object_id=identity_oid,
        target_object_id=target_oid,
        starts=starts,
        ends=ends,
    )
    db.commit()


def remove_permission(db, identity_key, target_oid, privilege, when=DEFAULT):
    identity_oid = key_lookup(db, identity_key)
    if when is DEFAULT:
        when = datetime.datetime.now()
    else:
        when = unstr_datetime(when)
    # base object is is the root to check for, user or group
    permission = db.permission
    query = permission.identity_object_id == identity_oid
    query &= permission.target_object_id == target_oid
    query &= permission.privilege == privilege
    query &= permission.starts <= when
    query &= permission.ends >= when
    result = db(query).delete() > 0
    db.commit()
    # print(db._lastsql)
    return result


def with_alias(db, source, alias):
    other = copy.copy(source)
    other['ALL'] = SQLALL(other)
    other['_tablename'] = alias
    for fieldname in other.fields:
        tmp = source[fieldname].clone()
        tmp.bind(other)
        other[fieldname] = tmp
    if 'id' in source and 'id' not in other.fields:
        other['id'] = other[source.id.name]

    if source_id := getattr(source, "_id", None):
        other._id = other[source_id.name]
    db[alias] = other
    return other


def has_permission(db, user_or_group_key, target_oid, privilege, when=DEFAULT):
    user_or_group_oid = key_lookup(db, user_or_group_key)
    # the permission system
    if when is DEFAULT:
        when = datetime.datetime.now()
    else:
        when = unstr_datetime(when)
    # base object is is the root to check for, user or group
    root_oid = user_or_group_oid
    permission = db.permission
    # ugly hack to satisfy pydal aliasing keyed tables /views
    left = with_alias(db, db.recursive_memberships, 'left')
    right = with_alias(db, db.recursive_memberships, 'right')
    # left = db.recursive_memberships.with_alias('left')
    # right = db.recursive_memberships.with_alias('right')

    # end of ugly hack
    query = left.root == root_oid
    query &= right.root == target_oid
    query &= permission.identity_object_id == left.object_id
    query &= permission.target_object_id == right.object_id
    query &= permission.privilege == privilege
    query &= permission.starts <= when
    query &= permission.ends >= when
    return db(query).count() > 0



from edwh_migrate import migration


@migration()
def rbac_migrations(db):
    db.executesql("""
    drop view if exists recursive_memberships;
    """)
    db.executesql("""
create view recursive_memberships as
  -- each root is member of object_id, including one line for himself.
  -- also for a user
  with RECURSIVE m(root, object_id, object_type, level, email, firstname, fullname) as (
        select object_id as root,  object_id, object_type, 0, email, firstname, fullname
          from identity
        union all
        select root, membership.member_of, i.object_type, m.level+1, i.email, i.firstname, i.fullname
          from membership join m on subject == m.object_id
               join identity i on i.object_id = membership.member_of
        order by root, m.level+1
    )
    select * from m
;
""")

    db.executesql("""
drop view if exists recursive_members;
    
    """)

    db.executesql("""
create view recursive_members as
    with RECURSIVE m(root, object_id, object_type, level, email, firstname, fullname) as (
        select object_id as root, object_id, object_type, 0, email, firstname, fullname
          from identity
        union all
        select root, membership.subject, i.object_type, m.level+1, i.email, i.firstname, i.fullname
          from membership join m on member_of== m.object_id
               join identity i on i.object_id = membership.subject
        order by root
    )
    select * from m
;

    """)

    db.commit()
    return True



import logging

from pydal import DAL
from . import model
from .model import Password, key_lookup, DEFAULT_STARTS, DEFAULT_ENDS, unstr_datetime, DEFAULT

_pylog = logging.getLogger(__name__)
_pylog.setLevel(logging.INFO)


class AuthRbac:
    name = 'auth_rbac'

    def __init__(self, db: DAL):
        self.db = db

    @staticmethod
    def _error(msg):
        print('ERROR:', msg)

    # gebruik event en rpc live templates (mobiel)

    def add_user(self, email, firstname, fullname, password, member_of):
        # check if exists
        email = email.lower().strip()
        user = model.get_user(self.db, email)
        if user:
            raise ValueError('User already exists')
        else:
            object_id = model.add_identity(self.db, email, password, member_of, firstname=firstname, fullname=fullname,
                                           object_type='user')
            rec = model.get_user(self.db, object_id)
            return dict(
                object_id=rec.object_id,
                email=rec.email,
                firstname=rec.firstname,
                fullname=rec.fullname,
            )

    def add_item(self, email, name, password, member_of):
        # check if exists
        email = email.lower().strip()
        item = model.get_identity(self.db, email, 'item')
        if item:
            raise ValueError('Item already exists')
        else:
            object_id = model.add_identity(self.db, email, name, None, password, member_of, object_type='item')
            rec = model.get_identity(self.db, object_id, object_type='item')
            return dict(
                object_id=rec.object_id,
                email=rec.email,
                name=rec.firstname
            )

    def add_identity(self, email, name, password, member_of, object_type):
        # check if exists
        email = email.lower().strip()
        if not isinstance(object_type, str):
            raise ValueError('object_type should be a string')
        identity = model.get_identity(self.db, email, object_type)
        if identity:
            raise ValueError('Item already exists')
        else:
            object_id = model.add_identity(self.db, email, name, None, password, member_of, object_type=object_type)
            rec = model.get_identity(self.db, object_id, object_type=object_type)
            return dict(
                object_id=rec.object_id,
                email=rec.email,
                name=rec.name
            )

    def add_group(self, email, name, member_of):
        # check if exists
        email = email.lower().strip()
        group = model.get_group(self.db, email)
        if group:
            raise ValueError('Group already exists')
        else:
            object_id = model.add_group(self.db, email, name, member_of)
            rec = model.get_group(self.db, object_id)
            return dict(
                object_id=rec.object_id,
                email=rec.email,
                name=rec.firstname
            )

    def update_identity(self, object_id, email=None, name=None, firstname=None, lastname=None, fullname=None,
                        password=None):
        user = model.get_identity(self.db, object_id)
        user.update_record(
            email=email.lower().strip() if email else user.email,
            firstname=name if name else firstname if firstname else user.firstname,
            lastname=lastname if lastname else user.lastname,
            fullname=fullname if fullname else user.fullname,
            password=Password.encode(password) if password else user.encoded_password,
        )
        self.db.commit()

    def get_user(self, key, return_memberhips=False):
        rec = model.get_user(self.db, key)
        result = dict(
            object_id=rec.object_id,
            email=rec.email,
            firstname=rec.firstname,
            fullname=rec.fullname,
        )
        if return_memberhips:
            result['memberships'] = [dict(
                object_id=m.object_id,
                object_type=m.object_type,
                email=m.email,
                firstname=m.firstname,
                fullname=m.fullname,
            ) for m in model.get_memberships(self.db, rec.object_id, bare=False)]
        return result

    def get_group(self, key, return_members=True):
        group_rec = model.get_group(self.db, key)
        members = []
        if return_members:
            members = model.get_members(self.db, group_rec.object_id, bare=False)
            members = [
                dict(
                    object_id=member.object_id,
                    object_type=member.object_type,
                    email=member.email,
                    name=member.firstname,
                )
                for member in members
            ]

        result = dict(
            object_id=group_rec.object_id,
            email=group_rec.email,
            name=group_rec.firstname
        )
        if return_members:
            result['members'] = members
        return result

    def authenticate_user(self, key, password):
        return model.authenticate_user(self.db, key=key, password=password)

    def add_membership(self, identity_key, group_key):
        return model.add_membership(self.db, identity_key, group_key)

    def remove_membership(self, identity_key, group_key):
        return model.remove_membership(self.db, identity_key, group_key)

    def has_membership(self, user_or_group_key, group_key):
        key = key_lookup(self.db, user_or_group_key)
        group = key_lookup(self.db, group_key)
        memberships = [m.object_id for m in model.get_memberships(self.db, key, bare=False)]
        return group in memberships

    def add_permission(self, identity_key, target_oid, privilege, starts=DEFAULT_STARTS, ends=DEFAULT_ENDS):
        starts = unstr_datetime(starts)
        ends = unstr_datetime(ends)
        model.add_permission(self.db, identity_key, target_oid, privilege, starts, ends)

    def has_permission(self, identity_key, target_oid, privilege, when=None):
        when = DEFAULT if when is None else unstr_datetime(when)
        return model.has_permission(self.db, identity_key, target_oid, privilege, when)

    def remove_permission(self, identity_key, target_oid, privilege, when=None):
        when = DEFAULT if when is None else unstr_datetime(when)
        return model.remove_permission(self.db, identity_key, target_oid, privilege, when=when)
